import sys
import templ8.blessing
import os
from templ8.divine import divine
from templ8.radio import radio
from templ8.genesis import genesis
from templ8.blessing import makedir
from templ8.pandoc import pandoc
from templ8.neo import neo
from templ8.help_args import help_cmd
import pkgutil

DEITY_PATH = "d8y"
DEF_BASEHTML_PATH = "basehtml"
DEF_REPLACE_PATH = "repl8ce"

DEF_INPUT = "input"
DEF_OUTPUT = "output"


def help(cmd=""):
	if cmd == "":
		print(templ8.blessing.TEMPL8_ASCII + "\n\n")
		print("Arguments in (parentheses) are optional. The ones in [brackets] are mandatory.\n")
		print("  help (command)   Display this list. If a command is given, displays instructions for that command.")
		print("  genesis [name]   Create a new templ8 project in a folder named [name].")
		print("  divine           Assemble a templ8 site.")
		print("  radio            Assemble a templ8 blog.")
		print("  neo              Upload all output files to Neocities.")
		print("  pandoc           Downloads a pandoc binary for markdown use.")
		print("")
	else:
		path = os.path.join("help", cmd)
		print("\n")
		print("=      == = ==  =  = ============ == = =   =         =")
		print("\n")
		print(help_cmd[cmd])
		print("\n")
		print("==     == ==== =  = ===== ======== == = =   ===      =")
		print("\n")

		


def main():
	if len(sys.argv) <= 1:
		help()
	elif sys.argv[1] == "help":
		if len(sys.argv) == 3:
			help(sys.argv[2])
		else:
			help()
	elif sys.argv[1] == "divine":
		if len(sys.argv) < 3:
			divine()
		elif sys.argv[2] == "--force":
			divine(True)
	elif sys.argv[1] == "radio":
		radio()
	elif sys.argv[1] == "pandoc":
		pandoc()
	elif sys.argv[1] == "neo":
		neo()
	elif sys.argv[1] == "genesis":
		if len(sys.argv) >= 3:
			genesis(sys.argv[2])
		else:
			raise Exception("Unexpected number of arguments")
	else:
		raise Exception("Unknown command")
	



import requests
from pathlib import Path
import os.path as pa
import os
import templ8.programstate as ps

def login():
	loginpath = Path.home() / ".templ8rc"
	lines = open(loginpath, "r").readlines()
	return lines[0].rstrip()
	

def neo():
	state = ps.ProgramState()
	last_change_list = {}
	if os.path.exists(".chlist"):
		for i in open(".chlist", "r").readlines():
			spl = i.split("<<")
			if len(spl) != 2:
				continue
			last_change_list[Path(spl[0])] = float(spl[1])
	new_change_list = last_change_list
	
	
	apikey = login()
	fls = []
	batch = 0
	finalprint = ""
	files_this_batch = 0
	for subdir, dirs, files in os.walk(state.output_folder):
		for file in files:
			if os.path.splitext(file)[1] != "":
				path = Path(pa.join(subdir, file))
				from_root = Path(*path.parts[1:])
				
				current_last_change = os.stat(path).st_mtime
				if path in last_change_list:
					if float(last_change_list[path]) == float(current_last_change):
						continue
				finalprint += f"{path}\n"
				new_change_list[path] = current_last_change
				if len(fls) <= batch:
					fls.append({})
				fls[batch][str(from_root).replace(os.sep, '/')] = open(path, "rb").read()
				files_this_batch += 1
				if files_this_batch > 50:
					finalprint += "\n\nNEW BATCH\n\n"
					files_this_batch = 0
					batch += 1
	endpoint = f"https://neocities.org/api/upload"
	new_change_text = ""
	for p in new_change_list:
		new_change_text += f"{p}<<{new_change_list[p]}\n"
	open(".chlist", "w").write(new_change_text)
	print(finalprint)
	for i in fls:
		response = requests.post(
			endpoint,
			headers={"Authorization": f"Bearer {apikey}"},
			files=i,
		)

		print("Response: ", response.status_code)
		print("Response: ", response.text)
	

			




import os
import shutil
import textile
from pathlib import Path
import templ8.programstate
from templ8.blessing import makedir
from templ8.blessing import mod_replaces
from templ8.blessing import full_parse
import time

# Divines a website
def divine(force=False):
	# Load the state of the program (important files and stuff)
	state = templ8.programstate.ProgramState()
	last_change_list = {}
	new_change_list = {}
	if os.path.exists(".chlistd"):
		for i in open(".chlistd", "r", encoding="utf-8").readlines():
			spl = i.split("<<")
			if len(spl) != 2:
				continue
			last_change_list[Path(spl[0])] = float(spl[1])
			new_change_list[Path(spl[0])] = float(spl[1])
	
	finalprint = ""
	for subdir, dirs, files in os.walk(state.input_folder):
		# Copy all the folders in case they're not there yet
		for dir in dirs:
			path = os.path.join(subdir, dir).replace(state.input_folder, state.output_folder, 1)
			makedir(path)
		
		dir_replace = {}
		dirpl8_ch = False
		# Find a global repl8ce thing
		if "repl8ce" in files:
			path = Path(os.path.join(subdir, "repl8ce"))
			mod_replaces(dir_replace, open(path, "r", encoding="utf-8").read())
			if path in last_change_list:
				if last_change_list[path] != os.stat(path).st_mtime:
					new_change_list[path] = os.stat(path).st_mtime
					dirpl8_ch = True
			else:
				new_change_list[path] = os.stat(path).st_mtime
					
		# Process the files
		for file in files:
			path = os.path.join(subdir, file)
			file_extension = os.path.splitext(path)[1]
			outpath = path.replace(state.input_folder, state.output_folder, 1)
			outhtml = outpath.replace(file_extension, ".html", -1)
			
			# Only process textile and md files
			if file_extension in [".textile", ".md", ".km"]:
				current_content = ""
				if os.path.exists(outhtml):
					current_content = open(outhtml, "r", encoding="utf-8").read()
						
				contents = ""
				# Get the headers and the content
				file_split = open(path, "r", encoding="utf-8").read().split("-BEGINFILE-",1)
				file_headers = ""
				file_content = ""
				if len(file_split) >= 2:
					file_headers = file_split[0]
					file_content = file_split[1]
				elif len(file_split) == 1:
					file_headers = ""
					file_content = file_split[0]
				else:
					raise Exception("Issue regarding -BEGINFILE- markers.")
				
				# Check if the folder is in txignore
				in_txignore = False
				for i in state.txignore:
					if os.path.join(state.input_folder, os.path.normpath(i)) == path or os.path.join(state.input_folder, os.path.normpath(i)) == subdir:
						in_txignore = True
						break
					
				
				if not in_txignore:
					npath = Path(path)
					origin_last_time = os.stat(npath).st_mtime
					recorded_last_time = -1
					if npath in last_change_list:
						recorded_last_time = last_change_list[npath]
					
					temp = state.replacements.copy()
					temp.update(dir_replace)
					mod_replaces(temp, file_headers)
					cbasech = False
					
					cbkpath = Path(state.basehtml_path)
					if "CUSTOMBASE" in temp:
						cbkpath = Path(temp["CUSTOMBASE"])
						
					if os.path.exists(cbkpath):
						if cbkpath in last_change_list:
							if last_change_list[cbkpath] != os.stat(cbkpath).st_mtime:
								new_change_list[cbkpath] = os.stat(cbkpath).st_mtime
								cbasech = True
						else:
							new_change_list[cbkpath] = os.stat(cbkpath).st_mtime
							cbasech = True
								
					if recorded_last_time != origin_last_time or cbasech or dirpl8_ch or force:
						contents = full_parse(state, file_content, file_extension, file_headers, dir_replace)
						new_change_list[npath] = origin_last_time
					else:
						continue
					
					if os.path.exists(outhtml):
						contents = contents.replace("#!CDATE!#", time.ctime(os.path.getctime(outhtml)))
					else:
						contents = contents.replace("#!CDATE!#", time.ctime(time.time()))
										
					finalprint += outhtml + "\n"
					with open(outhtml, "w", encoding="utf-8") as f:
						f.write(contents)
					
					
				else:
					finalprint += outpath + "\n"
					with open(outpath, "w", encoding="utf-8") as f:
						f.write(file_content)
					
			else:
				if os.path.exists(outpath):
					if open(path, "rb").read() == open(outpath, "rb").read():
						continue
				finalprint += outpath + "\n"
				shutil.copy(path, outpath)
		
	
	new_change_text = ""
	for p in new_change_list:
		new_change_text += f"{p}<<{new_change_list[p]}\n"
	open(".chlistd", "w", encoding="utf-8").write(new_change_text)
	
	print(finalprint)

	print("Finished assembling")



import os
import textile
import pypandoc
import time
import re
import pykami

pextra = ["-f", "markdown-smart"]

TEMPL8_ASCII = """
	###																	
	###																  ###
	###																  ###		#####
###########		   ###												  ###	  ##	###
###########		#########	   ### ####    #####	  ### ######	  ###	 ##		 ###
	###		   ###	   ###	   ###################	  ############	  ###	 ##		####
	###		  ###		###    ###	  ####	  ####	  ###	   ####   ###	  ##   ####
	###		  #############    ###	   ###	   ###	  ###		###   ###	   #######
	###		  #############    ###	   ###	   ###	  ###	   ####   ###	 ####	 ##
  #######	  ###			   ###	   ###	   ###	  ############	  ###	####	   ##
	###		  ####		###    ###	   ###	   ###	  ##########	  ###	###		   ##
	###			##########	   ###	   ###	   ###	  ###			  ###	 ###	 ###
	##			  ######	   ###	   ###	   ###	  ###			  ###	   ####### 
	##												  ###
	#"""
	
# Misnomer: This is the default blogbase
DEF_BASEHTML_CONTENT = """PAGETITLE=##TITLE##
-BEGINFILE-
h2. ##TITLE##

^##AUTHORS## - ##TAGS## - ##DATE##^

##CONTENT##

-BEGININDEX-
h2. "##TITLE##":##LINK##

^##AUTHORS## - ##TAGS## - ##DATE##^

##INTRO##
-BEGININDEX-
PAGETITLE=Blog"""

DEF_INPUT = "input"
DEF_OUTPUT = "output"
DEITY_PATH = ".d8y"
DEF_BASEHTML_PATH = "base.html"
DEF_REPLACE_PATH = ".repl8ce"


# Checks if a directory exists and make it if not
def makedir(path, warning = ""):
	if not os.path.exists(path):
		if warning != "":
			print("WARNING: " + warning)
		os.mkdir(path)


# Processes the repl8ce of a header
def mod_replaces(input_replaces, header):
	# These variables are used for multiline keys
	current_multikey = "" # The current key being edited
	multiline = False
	
	for i in header.split("\n"):
		# Split it in the single line way
		keyval = [i]
		if not multiline:
			keyval = i.split("=", 1)
		# If we're not processing multilines, and the current line is single line, set key
		if len(keyval) == 2 and not multiline:
			input_replaces[keyval[0]] = keyval[1].replace("\\n", "\n")
		# If it's not a valid single line
		elif len(keyval) == 1:
			# If it's a valid multiline, start a new multiline tag
			if keyval[0].startswith(";;"):
				multiline = True
				current_multikey = keyval[0].replace(";;", "", 1)
				input_replaces[current_multikey] = ""
			# If it's not, continue the current multiline
			elif multiline:
				input_replaces[current_multikey] += keyval[0] + "\n"
			elif not keyval[0] == "":
				raise Exception("what")
	
	for i in input_replaces:
		if i.startswith("TX-"):
			input_replaces[i] = parse_content(input_replaces[i], ".textile")
		elif i.startswith("MD-"):
			input_replaces[i] = parse_content(input_replaces[i], ".md")
		elif i.startswith("KM-"):
			input_replaces[i] = parse_content(input_replaces[i], ".km")


def parse_content(content, ext):
	if ext == ".textile":
		return textile.textile(content)
	elif ext == ".md":
		return pypandoc.convert_text(content, "html5", format="md", extra_args=pextra)
	elif ext == ".km":
		return pykami.parse(content)
	else:
		raise Exception("Can't recognize the extension in " + os.join(subdir, file))
		return ""

	
ifkey_start = "$$IF_"
ifnkey_start = "$$IF!"
forkey_start = "$$FOR_"
fkey_end = "$$END$$"


class Token:
	def __init__(self, type, start, end):
		self.end = end
		self.type = type
		self.start = start
	def text(self, input_base):
		return input_base[self.start:self.end]
	def __repr__(self):
		return f"[ C: {self.start} - {self.end} T: {self.type} ]"
	def __eq__(self, other):
		return (self.end == other.end and self.start == other.start)

tokens = [
	("if", re.compile(r"(?<!\\)\$ ?IF (NOT )?[A-Z0-9-_%]+")),
	("for", re.compile(r"(?<!\\)\$ ?FOR [A-Z0-9-_%]+ [A-Z]")),
	("end", re.compile(r"(?<!\\)\$ ?END")),
	("tag", re.compile(r"(?<!\\)\#\#[A-Z0-9-_%]+ ?(\#\#)?")),
]

openers = ["if", "for"]


def get_charpos(i, str):
	lines = str.splitlines()
	last_pos = 0
	for j in range(len(lines)):
		last_pos += len(lines[j])
		if last_pos > i:
			last_pos -= len(lines[j])
			return (j + 1, i-last_pos-j+1)


def parts(input_base):
	all_lexes = []
	last_end = 0
	while True:
		earliest = Token("nil", -1, -1)
		for name, gex in tokens:
			x = gex.search(input_base, last_end)
			if x and (x.span()[0] < earliest.start or earliest.type == "nil"):
				earliest = Token(name, x.span()[0], x.span()[1])
		
		if earliest.type != "nil":
			all_lexes.append(Token("put", last_end, earliest.start))
			all_lexes.append(earliest)	
			last_end = earliest.end
		else:
			all_lexes.append(Token("put", last_end, len(input_base)))
			break
	
	# Syntax error checking
	opens = []
	for i in all_lexes:
		if i.type in openers:
			opens.append(i)
		elif i.type == "end":
			if opens == []:
				errpos = get_charpos(i.start, input_base)
				print(f"ERROR: UNEXPECTED END OF BLOCK AT ({errppos[0]}; {errpos[1]})")
				close()
			else:
				opens.pop(-1)
	if opens != []:
		errpos = get_charpos(opens[-1].start, input_base)
		print(f"ERROR: UNCLOSED {opens[-1].type.upper()} AT ({errpos[0]}; {errpos[1]})")
		close()
	return all_lexes


def funkeys(input_base, keys, tokens, iter_variables = {}):
	out = ""
	iter = 0
	
	while iter < len(tokens):
		tok = tokens[iter]
		ttext = tok.text(input_base)
		
		# Replace all the %A iteration variables
		for i in iter_variables:
			ttext = ttext.replace("%"+i, str(iter_variables[i]))
			
		if tok.type == "put":
			out += ttext.replace("\\##", "##").replace("\\$", "$").replace("\\$", "$").replace("\n\n", "\n")
		if tok.type == "tag":
			tagley = ttext.replace("##", "").lstrip().rstrip()
			if tagley in keys and tagley != "CONTENT":
				out += keys[tagley]
		elif tok.type == "if":
			clex = ttext.replace("$IF", "").lstrip().rstrip()
			clex = clex.replace("$ IF", "").lstrip().rstrip()
			do = False
			if clex.startswith("NOT"):
				do = not do
				clex = clex.lstrip("NOT").lstrip()
			if clex in keys and keys[clex] != "":
				do = not do
			if not do:
				opens = 0
				for subtok in tokens[iter:]:
					if subtok.type in openers:
						opens += 1
					elif subtok.type == "end":
						opens -= 1
					if opens == 0:
						break
					iter += 1
		elif tok.type == "for":
			clex = ttext.replace("$ FOR", "").lstrip().rstrip()
			clex = clex.replace("$FOR", "").lstrip().rstrip()
			itervar = clex.split(" ")[1]
			clex = clex.split(" ")[0]
			opens = 0
			it2 = iter+1
			fit = 0
			for subtok in tokens[iter:]:
				if subtok.type in openers:
					opens += 1
				elif subtok.type == "end":
					opens -= 1
				if opens == 0:
					break
				it2 += 1
			while f"{clex}{fit}" in keys and keys[f"{clex}{fit}"] != "":
				out += funkeys(input_base, keys, tokens[iter+1:it2], {itervar: fit, **iter_variables})
				fit += 1
			iter = it2-1
		iter += 1
	return out


def into_html(content, keys, state):
	# By default, use the default base
	base = state.basehtml_content
	# Use a custombase if it's specified
	if "CUSTOMBASE" in keys:
		if os.path.exists(keys["CUSTOMBASE"]):
			base = open(keys["CUSTOMBASE"], "r").read()
		else:
			raise Exception(os.path.join(subdir, file) + " uses a CUSTOMBASE that doesn't exist")
	
	# Put the content in the file
	base = base.replace("##CONTENT##", content)
	
	# Replace special keys
	base = base.replace("#!DATE!#", time.ctime(time.time()))
	
	# Tokenize for function keys and apply them
	tokens = parts(base)
	if len(tokens) != 0:
		base = funkeys(base, keys, tokens)
	
	return base


def full_parse(state, file_content, file_extension, file_headers, dir_replace):
	# Process repl8ce
	filerepl = state.replacements.copy()
	filerepl.update(dir_replace)
	mod_replaces(filerepl, file_headers)
	
	# Turn the content into HTML
	contents = parse_content(file_content, file_extension)
	
	# Put the content in the base HTML
	contents = into_html(contents, filerepl, state)
	
	return contents


def parse_keys(page, keys):
	base = page
	tokens = parts(base)
	if len(tokens) != 0:
		base = funkeys(base, keys, tokens)
	
	for key in keys:
		if key.startswith("TX-"):
			base = base.replace("##"+key+"##", parse_content(keys[key], ".textile"))
		elif key.startswith("MD-"):
			base = base.replace("##"+key+"##", parse_content(keys[key], ".md"))
		else:
			base = base.replace("##"+key+"##", keys[key])
	
	return base
